# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 07.10.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Solution proposal for exercise 7, task 1.
"""

# pylint: disable=expression-not-assigned
# pylint: disable=bad-indentation

import os

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import cv2
import seaborn as sns # pylint: disable=import-error,unused-import
import scipy.io # pylint: disable = import-error

import multivariate_gauss_classifier

def plot_image(image, fig_num, name=None, colormap='gray', write_file=None):
  """Plot image"""
  plt.figure(fig_num)
  plt.imshow(image, cmap=colormap, interpolation='none')
  if name is not None:
    plt.title(name)
  plt.xticks([]), plt.yticks([])
  plt.tight_layout()
  if write_file:
    cv2.imwrite(write_file, image)
  fig_num += 1
  return fig_num

def plot_class_image(image, fig_num, classes=None, name=None, colormap='gray',
                     write_file=None):
  """Plot class image"""
  plt.figure(fig_num)
  if classes is None:
    classes = np.unique(image)

  if not isinstance(classes, list):
    classes = list(classes)

  cmap = plt.cm.get_cmap(colormap)

  patches = []
  for c in classes:
    color = cmap(c/np.max(classes))
    classname = 'C '+str(c)
    patches.append(mpatches.Patch(color=color, label=classname))

  plt.imshow(image, cmap=colormap, interpolation='none', vmin=0, vmax=classes[-1])
  if name is not None:
    plt.title(name)
  plt.xticks([]), plt.yticks([])
  plt.legend(bbox_to_anchor=(0, 1.02, 1., .102), loc=3, ncol=len(classes),
             mode="expand", borderaxespad=0, handles=patches)
  if write_file:
    cv2.imwrite(write_file, image)
  fig_num += 1
  return fig_num

def plot_posteriors(posteriors, fig_num, name=None, colormap='gray'):
  """Plot posterior images (assuming 4 classes)"""

  fig = plt.figure(fig_num)
  ax1 = fig.add_subplot(221)
  ax2 = fig.add_subplot(222)
  ax3 = fig.add_subplot(223)
  ax4 = fig.add_subplot(224)

  if name is not None:
    fig.suptitle(name)

  minval = np.min(posteriors)
  maxval = np.max(posteriors)

  ax1.imshow(posteriors[0, :, :], vmin=minval, vmax=maxval,
             cmap=colormap, interpolation='none')
  ax1.set_xticks([]), ax1.set_yticks([])
  ax1.set_title('Class 1')

  ax2.imshow(posteriors[1, :, :], vmin=minval, vmax=maxval,
             cmap=colormap, interpolation='none')
  ax2.set_xticks([]), ax2.set_yticks([])
  ax2.set_title('Class 2')

  ax3.imshow(posteriors[2, :, :], vmin=minval, vmax=maxval,
             cmap=colormap, interpolation='none')
  ax3.set_xticks([]), ax3.set_yticks([])
  ax3.set_title('Class 3')

  im = ax4.imshow(posteriors[3, :, :], vmin=minval, vmax=maxval,
                  cmap=colormap, interpolation='none')
  ax4.set_xticks([]), ax4.set_yticks([])
  ax4.set_title('Class 4')

  fig.subplots_adjust(right=0.8)
  cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
  fig.colorbar(im, cax=cbar_ax)

  #plt.tight_layout()

  fig_num += 1
  return fig_num

def plot_feature_distributions(class_feature_values, fig_num, classes,
                               colormap='rainbow'):
  """
  Plot the distribution of the feature values for each class (from where we
  estimate the class mean and class variance
  """

  plt.figure(fig_num)
  cmap = plt.cm.get_cmap(colormap)

  minval = min([np.min(feature_vals) for feature_vals in class_feature_values])
  maxval = max([np.max(feature_vals) for feature_vals in class_feature_values])
  bins = np.linspace(minval, maxval, int(np.ceil(maxval) - np.floor(minval)))
  for i, c in enumerate(classes):
    color = cmap(c/np.max(classes))
    plt.hist(class_feature_values[i], bins=bins, color=color,
             label='C '+str(c), histtype='step', linewidth=5)
    #sns.kdeplot(class_feature_values[i], label='C '+str(c),
    #            c=color, kernel='triw')
  plt.legend(loc='upper right')

  fig_num += 1
  return fig_num

def main():
  """Classifying images using a multivariate gaussian classifier"""

  print('='*80)
  print('Solution to weekly exercises in INF4300')
  print('Exercise 7')
  print('Task 1')
  print('-'*80)

  # Configurations
  image_dir = '../../images'
  fig_num = 0
  iid = True # Assuming independent features with equal variances
  class_colormap = 'rainbow'
  train = True
  classify = True
  evaluate = True
  feature_filenames = ['tm1.png',
                       'tm2.png',
                       'tm3.png',
                       'tm4.png',
                       'tm5.png',
                       'tm6.png'
                      ]

  if train:
    train_mask = cv2.imread(os.path.join(image_dir, 'tm_train.png'), cv2.IMREAD_GRAYSCALE)
    fig_num = plot_class_image(train_mask, fig_num, colormap=class_colormap)
    M, N = train_mask.shape

  if evaluate:
    test_mask = cv2.imread(os.path.join(image_dir, 'tm_test.png'), cv2.IMREAD_GRAYSCALE)
    fig_num = plot_class_image(test_mask, fig_num, colormap=class_colormap)

  feature_images = []
  for fname in feature_filenames:
    print('Using feature {}'.format(fname))
    feature_images.append(cv2.imread(os.path.join(image_dir, fname), cv2.IMREAD_GRAYSCALE))
    #fig_num = plot_image(feature_images[-1], fig_num)

  mgc = multivariate_gauss_classifier.Classification()

  if train:
    mgc.train(train_mask, feature_images, iid)
    classes = mgc.classes
    class_means = mgc.class_means
    class_covariances = mgc.class_covariances

    print('Finished training')

  if classify:
    mgc.classify(feature_images, class_means, class_covariances)
    posterior_images = mgc.posterior_images
    prop_classified_image = mgc.prop_classified_image

    print('Finished segmentation')

    fig_num = plot_class_image(prop_classified_image, fig_num,
                               classes=[0] + [c for c in classes],
                               colormap=class_colormap)

    fig_num = plot_posteriors(posterior_images, fig_num, colormap='magma')

  if evaluate:
    mgc.evaluate(test_mask, prop_classified_image)
    confusion_matrix = mgc.confusion_matrix
    mgc.print_evaluation_measures(confusion_matrix)

    # Check with external true result

    # Load .mat files in as a dictionary where one of the elements is the data
    # from the image
    # print(tm_test_result.keys()) # For info about what this mat file contains
    # 'klassim' is a (260, 333) double image, a true result so we can test our implementation
    tm_test_result = scipy.io.loadmat('data/tm_classres.mat')
    num_equal = np.sum(1*(prop_classified_image == tm_test_result['klassim']))

    print('Agreement with provided classified image: {0:.2f} %'.format(num_equal*100/(M*N)))

    print('Finished evaluation')



  plt.show()

if __name__ == "__main__":
  main()
